#include <stdbool.h>
#include <stdlib.h>

#include "ki/dynarray.h"

Dynarray* DynarrayNew(size_t element_size) {
    Dynarray *darray = ki_malloc(sizeof(Dynarray));

    if (!darray) {
        return NULL;
    }

    darray->elements     = NULL;
    darray->element_size = element_size;
    darray->len          = 0;

    return darray;
}

Dynarray* DynarrayNewLength(size_t element_size, size_t length) {
    Dynarray *darray = ki_malloc(sizeof(Dynarray));

    if (!darray) {
        return NULL;
    }

    if (length > 0) {
        darray->elements = ki_malloc(element_size * length);

        if (!darray->elements) {
            free(darray);
            return NULL;
        }
    }

    darray->element_size = element_size;
    darray->len          = length;

    return darray;
}

bool DynarrayEnsureCapacity(Dynarray *darray, size_t length) {
    void *new_elements = NULL;

    if (darray->len >= length) {
        return true;
    }

    new_elements = ki_realloc(darray->elements, length * darray->element_size);

    if (!new_elements) {
        return false;
    }

    darray->elements = new_elements;
    darray->len      = length;

    return true;
}

void* DynarrayNewElementAtEnd(Dynarray *darray) {
    if (!DynarrayEnsureCapacity(darray, darray->len + 1)) {
        return NULL;
    }

    return DynarrayIndex(darray, darray->len - 1);
}

void* DynarrayNewElementAtBeginning(Dynarray *darray) {
    char *elements = NULL;

    if (!DynarrayEnsureCapacity(darray, darray->len + 1)) {
        return NULL;
    }

    elements = (char *)darray->elements;

    memmove(
        elements,
        elements + darray->element_size,
        (darray->len - 1) * darray->element_size
    );

    memset(elements, 0, darray->element_size);

    return DynarrayIndex(darray, 0);
}

void* DynarrayIndex(Dynarray *darray, size_t index) {
    if (index >= darray->len) {
        return NULL;
    }

    return ((char *)darray->elements) + (index * darray->element_size);
}

void DynarrayClear(Dynarray *darray) {
    memset(darray->elements, 0, darray->len * darray->element_size);
}

void DynarrayFree(Dynarray *darray) {
    ki_free(darray->elements);
    darray->element_size = 0;
    darray->len          = 0;
    darray->elements     = NULL;
}

/* vi: set et ts=2 sw=2: */


